require 'spec_helper'

class DummyEntity
  include Entity

  attr_accessor :name

  def set_name(name)
    record_event DummyEntityNameSet.new(name: name)
  end
end

class DummyEntityNameSet
  include Event

  attr_accessor :name

  def apply(entity)
    entity.name = name
  end
end

class DummyEntitySubscriber
  class << self
    attr_accessor :event_name
  end

  def dummy_entity_name_set(event)
    DummyEntitySubscriber.event_name = event.name
  end
end

describe "end to end" do
  before(:each) do
    PostgresEntityStore.connection_string = ENV['SQL_CONNECTION'] || 'postgres://localhost/cronofy_test'
    PostgresEntityStore.init

    EntityStore::Config.setup do |config|
      config.store = PostgresEntityStore.new
      config.event_subscribers << DummyEntitySubscriber
    end
  end

  context "when save entity" do
    let(:name) { random_string }
    before(:each) do
      @entity = DummyEntity.new
      @entity.set_name name
      @id = Store.new.save @entity
    end

    it "publishes event to the subscriber" do
      expect(DummyEntitySubscriber.event_name).to eq(name)
    end
    it "is retrievable with the events applied" do
      expect(EntityStore::Store.new.get(@entity.id).name).to eq(name)
    end
  end
end

describe "setting connection per instance" do
  before(:each) do
    PostgresEntityStore.connection_string = 'FAKE_CONNECTION_STRING'

    EntityStore::Config.setup do |config|
      config.store = PostgresEntityStore.new(Sequel.connect('postgres://localhost/cronofy_test'))
      config.event_subscribers << DummyEntitySubscriber
    end

    PostgresEntityStore.init
  end

  context "when save entity" do
    let(:name) { random_string }
    before(:each) do
      @entity = DummyEntity.new
      @entity.set_name name
      @id = Store.new.save @entity
    end

    it "publishes event to the subscriber" do
      expect(DummyEntitySubscriber.event_name).to eq(name)
    end
    it "is retrievable with the events applied" do
      expect(EntityStore::Store.new.get(@entity.id).name).to eq(name)
    end
  end
end

describe "setting connection from existing" do
  before(:each) do
    PostgresEntityStore.database = Sequel.connect('postgres://localhost/cronofy_test')
    PostgresEntityStore.init

    EntityStore::Config.setup do |config|
      config.store = PostgresEntityStore.new
      config.event_subscribers << DummyEntitySubscriber
    end
  end

  context "when save entity" do
    let(:name) { random_string }
    before(:each) do
      @entity = DummyEntity.new
      @entity.set_name name
      @id = Store.new.save @entity
    end

    it "publishes event to the subscriber" do
      expect(DummyEntitySubscriber.event_name).to eq(name)
    end
    it "is retrievable with the events applied" do
      expect(EntityStore::Store.new.get(@entity.id).name).to eq(name)
    end
  end
end
